@@ -5,31 +5,79 @@ use vars qw($VERSION);
 use DB_File;
 use POSIX;
 
-$VERSION = '0.06';
+$VERSION = '0.07';
 
 use constant PI => 3.14159265;
-use constant EARTH_RADIUS => 3956;
-use constant EARTH_CIRCUMFERENCE => 2 * PI * EARTH_RADIUS;
-use constant LON_MILEAGE_CONSTANT => 69.1716;
 use constant LAT_DEGREES => 180;
 use constant LON_DEGREES => 360;
-use constant LAT_WIDTH => (PI * EARTH_RADIUS / LAT_DEGREES);
+
+# Earth radius in various units
+our %_UNITS = (
+	       mi    => 3956,
+	       km    => 6376.5,
+	      );
+# Aliases
+$_UNITS{$_} = $_UNITS{mi}
+  foreach (qw(mile miles));
+$_UNITS{$_} = $_UNITS{km}
+  foreach (qw(kilometer kilometers));
+
+sub _max
+{
+  my $max = shift;
+  foreach my $v (@_)
+  {
+    $max = $v
+      if ($v > $max);
+  }
+  $max;
+}
 
 sub new {
   my ($class, %options) = @_;
-  my (%postalcode, %city, %latlon);
   my $db_dir = $options{db_dir};
-  tie %postalcode, 'DB_File', "$db_dir/postalcode.db", O_CREAT|O_RDWR, 0666, $DB_BTREE;
-  tie %city,       'DB_File', "$db_dir/city.db",       O_CREAT|O_RDWR, 0666, $DB_BTREE;
-  tie %latlon,     'DB_File', "$db_dir/latlon.db",     O_CREAT|O_RDWR, 0666, $DB_BTREE;
-  bless {postalcode => \%postalcode, city => \%city, latlon => \%latlon}, $class;
+  my $btree;
+  if ($options{cachesize}) {
+    if ($options{cachesize} eq 'max') {
+      $options{cachesize} = _max(map { -s "$db_dir/$_.db"} qw(postalcode city latlon));
+    }
+    $btree = new DB_File::BTREEINFO;
+    $btree->{cachesize} = $options{cachesize};
+  } else {
+    $btree = $DB_BTREE;
+  }
+
+  my (%postalcode, %city, %latlon);
+
+  tie %postalcode, 'DB_File', "$db_dir/postalcode.db", O_RDONLY, 0666, $btree
+      or die "Couldn't open postalcode.db in $db_dir: $!\n";
+  tie %city,       'DB_File', "$db_dir/city.db",       O_RDONLY, 0666, $btree
+      or die "Couldn't open city.db in $db_dir: $!\n";
+  tie %latlon,     'DB_File', "$db_dir/latlon.db",     O_RDONLY, 0666, $btree
+      or die "Couldn't open latlon.db in $db_dir: $!\n";
+  my $self = {postalcode => \%postalcode, city => \%city, latlon => \%latlon};
+
+  if ($options{units} && $_UNITS{lc $options{units}})
+  {
+    $self->{_earth_radius}=$_UNITS{lc $options{units}}
+  }
+  elsif ($options{earth_radius})
+  {
+    $self->{_earth_radius} = $options{earth_radius};
+  }
+  else
+  {
+    $self->{_earth_radius} = $_UNITS{mi};
+  }
+  
+  bless $self, $class;
 }
 
 sub lookup_postal_code {
   my ($self, %options) = @_;
   my $v = $self->{postalcode}->{$options{postal_code}};
   return unless $v;
-  my ($lat, $lon, $city, $state) = split(",",$v);
+  my ($lat, $lon, $city, $state) = split(/,/,$v);
   return {lat => $lat, lon => $lon, city => $city, state => $state};
 }
 
@@ -38,7 +86,7 @@ sub lookup_city_state {
   my $city_state = uc(join("", $options{state}, $options{city}));
   my $v = $self->{city}->{$city_state};
   return unless $v;
-  my ($postal_code_str, $lat, $lon) = split('\|',$v);
+  my ($postal_code_str, $lat, $lon) = split(/\|/,$v);
   my @postal_codes = ($postal_code_str =~ m!(.{5})!g);
   return {lat => $lat, lon => $lon, postal_codes => \@postal_codes};
 }
@@ -49,12 +97,13 @@ sub calculate_distance {
   my $ra = $self->lookup_postal_code(postal_code => $a);
   my $rb = $self->lookup_postal_code(postal_code => $b);
   return unless $ra && $rb;
-  return _calculate_distance($ra->{lat}, $ra->{lon}, $rb->{lat}, $rb->{lon});
+  return _calculate_distance($ra->{lat}, $ra->{lon}, $rb->{lat}, $rb->{lon}, $self->{_earth_radius});
 }
 
+# in miles
 # in miles
 sub _calculate_distance {
-  my ($lat_1, $lon_1, $lat_2, $lon_2) = @_;
+  my ($lat_1, $lon_1, $lat_2, $lon_2, $rho) = @_;
 
   # Convert all the degrees to radians
   $lat_1 *= PI/180;
@@ -69,7 +118,7 @@ sub _calculate_distance {
   # Find the Great Circle distance
   my $temp = sin($delta_lat/2.0)**2 + cos($lat_1) * cos($lat_2) * sin($delta_lon/2.0)**2;
 
-  return EARTH_RADIUS * 2 * atan2(sqrt($temp),sqrt(1-$temp));
+  return $rho * 2 * atan2(sqrt($temp),sqrt(1-$temp));
 }
 
 sub nearby_postal_codes {
@@ -79,10 +128,12 @@ sub nearby_postal_codes {
 
 sub query_postal_codes {
   my ($self, %options) = @_;
+  my $pcdb = $self->{postalcode};
+  my $lldb = $self->{latlon};
   my ($lat, $lon, $distance, $order_by) = @options{qw(lat lon distance order_by)};
   my %select = map {$_ => 1} @{$options{select}};
 
-  my $distance_degrees = _min($distance/LAT_WIDTH,
+  my $distance_degrees = _min($distance/(PI * $self->{_earth_radius} / LAT_DEGREES),
 			      LAT_DEGREES);
   my $min_lat = floor($lat - $distance_degrees);
   my $max_lat = floor($lat + $distance_degrees);
@@ -105,7 +156,6 @@ sub query_postal_codes {
     }
     elsif ($x < -(LAT_DEGREES/2))
     {
-      my $w = "Changed latitude: $x -> ";
       $x = -$x - LAT_DEGREES;
       $lon_rtw = 1;
     }
@@ -122,7 +172,7 @@ sub query_postal_codes {
     }
     else
     {
-      $lon_distance_degrees = _min($distance / _min(_lon_miles($x),_lon_miles($x+1)),
+      $lon_distance_degrees = _min($distance / _min($self->_lon_miles($x),$self->_lon_miles($x+1)),
 				   LON_DEGREES/2);
     }
 
@@ -147,7 +197,6 @@ sub query_postal_codes {
     }
 
     for my $y ($min_lon .. $max_lon) {
-
       # Correct longitude for latitude that wrapped 'round-the-world.
       if ($lon_rtw)
       {
@@ -166,26 +215,34 @@ sub query_postal_codes {
 
       next unless _calculate_distance($lat, $lon,
 				      _test_near($lat, $x),
-				      _test_near($lon, $y)) <= $distance;
-      my $postal_code_str = $self->{latlon}->{"$x-$y"};
+				      _test_near($lon, $y),
+				      $self->{_earth_radius}) <= $distance;
+      my $postal_code_str = $lldb->{"$x-$y"};
       next unless $postal_code_str;
       my @cell_zips = ($postal_code_str =~ m!(.{5})!g);
       if (_calculate_distance($lat, $lon,
 			      _test_far($lat, $x),
-			      _test_far($lon, $y)) <= $distance) {
+			      _test_far($lon, $y),
+			      $self->{_earth_radius}) <= $distance) {
 	# include all of cell
 	for (@cell_zips) {
 	  my %h = (postal_code => $_);
 	  if ($select{distance}||$select{lat}||$select{lon}||$select{city}||$select{state})
 	  {
-	    my $r = $self->lookup_postal_code(postal_code => $_);
-	    next unless $r;
+	    my($rlat,$rlon,undef)=split(/,/,$pcdb->{$_},3);
+	    my $r;
 	    for my $field (keys %select) {
 	      if ($field eq 'distance') {
-		$h{distance} = _calculate_distance($lat,$lon,$r->{lat},$r->{lon});
+		$h{distance} = _calculate_distance($lat,$lon,$rlat,$rlon,$self->{_earth_radius});
 	      } elsif ($field eq 'postal_code') {
 		; # Do Nothing.
-	      } else {
+	      } elsif ($field eq 'lat') {
+		$h{lat} = $rlat;
+	      } elsif ($field eq 'lon') {
+		$h{lon} = $rlon;
+  	      } else {
+                $r = $self->lookup_postal_code(postal_code => $_)
+		    unless $r;
 		$h{$field} = $r->{$field};
 	      }
 	    }
@@ -195,9 +252,10 @@ sub query_postal_codes {
       } else {
 	# include only postal code with distance
 	for (@cell_zips) {
-	  my $r = $self->lookup_postal_code(postal_code => $_);
-	  next unless $r;
-	  my $d = _calculate_distance($lat, $lon, $r->{lat}, $r->{lon});
+	  # Can we guarantee this will never be undef?...
+	  my($rlat,$rlon,undef)=split(/,/,$pcdb->{$_},3);
+          my $r;
+	  my $d = _calculate_distance($lat, $lon, $rlat, $rlon,$self->{_earth_radius});
 	  if ($d <= $distance) { 
 	    my %h = (postal_code => $_);
 	    for my $field (keys %select) {
@@ -205,7 +263,14 @@ sub query_postal_codes {
 		$h{distance} = $d;
 	      } elsif ($field eq 'postal_code') {
 		; # Do Nothing.
-	      } else {
+	      } elsif ($field eq 'lat') {
+		$h{lat} = $rlat;
+	      } elsif ($field eq 'lon') {
+		$h{lon} = $rlon;
+	      }
+	      else { 
+                $r = $self->lookup_postal_code(postal_code => $_)
+		    unless $r;
 		$h{$field} = $r->{$field};
 	      }
 	    }
@@ -258,11 +323,12 @@ sub _test_far {
 
 sub _lon_miles
 {
+  my $self = shift;
   my($lat)=@_;
 
   # Formula from:
   #   http://www.malaysiagis.com/related_technologies/mapping/basics1b.cfm
-  my $r = cos($lat*PI/180)*LON_MILEAGE_CONSTANT;
+  my $r = cos($lat*PI/180)*(2*PI*$self->{_earth_radius}/LON_DEGREES);
   $r;
 }
 
@@ -295,7 +361,7 @@ Geo::PostalCode - Find closest zipcodes, distance, latitude, and longitude.
   my $record = $gp->lookup_city_state(city => "Jersey City",state => "NJ");
   my $lat          = $record->{lat};
   my $lon          = $record->{lon};
-  my $postal_codes = $record->{postal_codes}:
+  my $postal_codes = $record->{postal_codes};
 
   my $postal_codes = $gp->nearby_postal_codes(lat => $lat, lon => $lon,
                                                    distance => 50);
@@ -317,11 +383,27 @@ To access the data, it uses Berkeley DB, which is fast and portable
 
 =over 4
 
-=item $gp = Geo::PostalCode->new(db_dir => $db_dir);
+=item $gp = Geo::PostalCode->new(db_dir => $db_dir,
+                                 [cachesize => btree_cache_size | max ,]
+                                 [units => mi | km ,]
+                                 [earth_radius => earth_radius_in_desired_units ,]
+                                );
 
 Returns a new Geo::PostalCode object using the postalcode.db, latlon.db, and city.db
 Berkeley Database files in $db_dir.
 
+You can override the cache used by the BTree in L<DB_File|DB_File> by
+setting the C<cachesize> parameter to the desired size of the cache,
+in bytes.  You can set it to C<max> to set it to the size of the
+largest of the databases used.
+
+You can control the distance units used by providing a C<units>
+option, which can be C<mi> for miles (the default) or C<km> for
+kilometers, or by providing a C<earth_radius> option set to the radius
+of the Earth in your desired unit.  The Earth's radius is
+approximately 3956 miles.
+
+
 =item $record = $gp->lookup_postal_code(postal_code => $postal_code);
 
 Returns a hash reference containing four keys:
@@ -361,6 +443,7 @@ Each hash reference contains the following fields:
 
 If $order_by is specified, then the records are sorted by the $order_by field.
 
+
 =back
 
 =head1 NOTES
@@ -375,20 +458,20 @@ It is available from http://www.cryptnet.net/fsp/zipdy/
 
 =head1 AUTHOR
 
-Copyright (c) 2002, T.J. Mather, tjmather@tjmather.com
+Copyright (c) 2006, MaxMind LLC, http://www.maxmind.com/
 
 All rights reserved.  This package is free software; you can redistribute it
 and/or modify it under the same terms as Perl itself.
 
 =head1 CREDITS
 
-Thanks to Scott Gifford for contributing multiple bug fixes and code cleanup.
+Thanks to Scott Gifford of http://homesurfusa.com/ for contributing multiple bug fixes and code cleanup.
 
 =head1 SEE ALSO
 
 =over 4
 
-L<Geo::IP> - Look up country by IP Address
+L<Geo::IP> - Look up country and city by IP Address
 
 zipdy - Free Zip Code Distance Calculator
 
